module Morsi (morsiCodeFromFreqs) where
import Data.Function
import qualified Data.List
import qualified Data.Maybe

{-WETT-}
data HuffTree a =
    Leaf Integer a |
    Inner Integer (HuffTree a) (HuffTree a)
  deriving (Show)

weight :: HuffTree a -> Integer
weight (Leaf w _) = w
weight (Inner w _ _) = w

insortTree :: HuffTree a -> [HuffTree a] -> [HuffTree a]
insortTree t [] = [t]
insortTree t (t' : ts) =
  if weight t <= weight t' then t : t' : ts else t' : insortTree t ts

mergeTrees :: HuffTree a -> HuffTree a -> HuffTree a
mergeTrees t t' = Inner (weight t + weight t') t t'

huffman :: [HuffTree a] -> HuffTree a
huffman [t] = t
huffman (t : t' : ts) = huffman (insortTree (mergeTrees t t') ts)

huffmanTreeFromFreqs :: [(a, Integer)] -> HuffTree a
huffmanTreeFromFreqs =
  huffman . map (uncurry (flip Leaf)) . Data.List.sortBy (compare `on` snd)

huffmanCodeFromTree :: Char -> Char -> HuffTree a -> [(a, String)]
huffmanCodeFromTree l r = code []
  where code pref (Leaf _ a) = [(a, pref)]
        code pref (Inner _ t t') = code (pref ++ [l]) t ++ code (pref ++ [r]) t'

morsiCodeFromTree = huffmanCodeFromTree 'A' 'i'

morsiCodeFromFreqs :: [(a, Integer)] -> [(a, String)]
morsiCodeFromFreqs = morsiCodeFromTree . huffmanTreeFromFreqs

costOfMorsiSym :: Char -> Integer
costOfMorsiSym 'i' = 1
costOfMorsiSym 'A' = 2

costOfMorsiWord :: String -> Integer
costOfMorsiWord = sum . map costOfMorsiSym

costOfMorsiCode :: Eq a => [(a, Integer)] -> Integer
costOfMorsiCode freqs =
  sum costs
  where code = morsiCodeFromFreqs freqs
        isas = map (\(a, bs) -> (a, costOfMorsiWord bs)) code
        costs = [isa * freq | (a, freq) <- freqs, (a', isa) <- isas, a == a']

encodeSym :: Eq a => [(a, String)] -> a -> String
encodeSym code a = Data.Maybe.fromJust (Data.List.lookup a code)

encodeWord :: Eq a => [(a, String)] -> [a] -> String
encodeWord = concatMap . encodeSym

decodeSym :: [(a, String)] -> String -> (a, String)
decodeSym code (b : bs) =
  case Data.List.find (null . snd) code' of
    Just (a, _) -> (a, bs)
    Nothing -> decodeSym code' bs
  where code' = [(a, cs) | (a, c : cs) <- code, c == b]

decodeWord :: [(a, String)] -> String -> [a]
decodeWord _ [] = []
decodeWord code bs =
  a : decodeWord code bs'
  where (a, bs') = decodeSym code bs
{-TTEW-}

zdfFreqs = [('z', 2), ('d', 3), ('f', 5), ('s', 7), ('e', 11), ('h', 100)]
zdfCode = morsiCodeFromFreqs zdfFreqs

zdfCost = costOfMorsiCode zdfFreqs

zdfOrig = "zdfsezdfsedfsefsefseseseeeee"
zdfCoded = encodeWord zdfCode zdfOrig
zdfDecoded = decodeWord zdfCode zdfCoded
